@@ -22,11 +22,13 @@
 #include "config.h"
 #endif
 
+#include <stdio.h>
 #include <string.h>
 
 #include <gst/gst.h>
 #include <gst/tag/tag.h>
 #include <gst/riff/riff-media.h>
+#include <gst/riff/riff-read.h>
 
 GST_DEBUG_CATEGORY_STATIC (gst_ogm_parse_debug);
 #define GST_CAT_DEFAULT gst_ogm_parse_debug
@@ -59,19 +61,19 @@ static const GstElementDetails gst_ogm_audio_parse_details =
 GST_ELEMENT_DETAILS ("OGM audio stream parser",
     "Codec/Decoder/Audio",
     "parse an OGM audio header and stream",
-    "Ronald Bultje <rbultje@ronald.bitfreak.net>");
+    "GStreamer maintainers <gstreamer-devel@lists.sourceforge.net>");
 
 static const GstElementDetails gst_ogm_video_parse_details =
 GST_ELEMENT_DETAILS ("OGM video stream parser",
     "Codec/Decoder/Video",
     "parse an OGM video header and stream",
-    "Ronald Bultje <rbultje@ronald.bitfreak.net>");
+    "GStreamer maintainers <gstreamer-devel@lists.sourceforge.net>");
 
 static const GstElementDetails gst_ogm_text_parse_details =
 GST_ELEMENT_DETAILS ("OGM text stream parser",
     "Codec/Decoder/Subtitle",
     "parse an OGM text header and stream",
-    "Ronald Bultje <rbultje@ronald.bitfreak.net>");
+    "GStreamer maintainers <gstreamer-devel@lists.sourceforge.net>");
 
 typedef struct _stream_header_video
 {
@@ -230,7 +232,7 @@ gst_ogm_audio_parse_get_type (void)
   return ogm_audio_parse_type;
 }
 
-GType
+static GType
 gst_ogm_video_parse_get_type (void)
 {
   static GType ogm_video_parse_type = 0;
@@ -256,7 +258,7 @@ gst_ogm_video_parse_get_type (void)
   return ogm_video_parse_type;
 }
 
-GType
+static GType
 gst_ogm_text_parse_get_type (void)
 {
   static GType ogm_text_parse_type = 0;
@@ -570,12 +572,13 @@ gst_ogm_parse_stream_header (GstOgmParse * ogm, const guint8 * data, guint size)
       GST_LOG_OBJECT (ogm, "Type: %s, subtype: 0x%04x, channels: %d, "
           "samplerate: %d, blockalign: %d, bps: %d, caps = %" GST_PTR_FORMAT,
           ogm->hdr.streamtype, codec_id, ogm->hdr.s.audio.channels,
-          ogm->hdr.samples_per_unit, ogm->hdr.s.audio.blockalign,
+          (gint) ogm->hdr.samples_per_unit, ogm->hdr.s.audio.blockalign,
           ogm->hdr.s.audio.avgbytespersec, caps);
       break;
     }
     case 'v':{
       guint32 fourcc;
+      gint time_unit;
 
       fourcc = GST_MAKE_FOURCC (ogm->hdr.subtype[0],
           ogm->hdr.subtype[1], ogm->hdr.subtype[2], ogm->hdr.subtype[3]);
@@ -600,10 +603,15 @@ gst_ogm_parse_stream_header (GstOgmParse * ogm, const guint8 * data, guint size)
           ogm->hdr.samples_per_unit, ogm->hdr.default_len,
           ogm->hdr.buffersize, ogm->hdr.bits_per_sample, caps);
 
+      /* GST_TYPE_FRACTION contains gint */
+      if (ogm->hdr.time_unit > G_MAXINT || ogm->hdr.time_unit < G_MININT)
+        GST_WARNING_OBJECT (ogm, "timeunit is out of range");
+
+      time_unit = (gint) CLAMP (ogm->hdr.time_unit, G_MININT, G_MAXINT);
       gst_caps_set_simple (caps,
           "width", G_TYPE_INT, ogm->hdr.s.video.width,
           "height", G_TYPE_INT, ogm->hdr.s.video.height,
-          "framerate", GST_TYPE_FRACTION, 10000000, ogm->hdr.time_unit, NULL);
+          "framerate", GST_TYPE_FRACTION, 10000000, time_unit, NULL);
       break;
     }
     case 't':{
@@ -626,6 +634,7 @@ gst_ogm_parse_stream_header (GstOgmParse * ogm, const guint8 * data, guint size)
     if (current_caps && caps && !gst_caps_is_equal (current_caps, caps)) {
       GST_WARNING_OBJECT (ogm, "Already an existing pad %s:%s",
           GST_DEBUG_PAD_NAME (ogm->srcpad));
+      gst_pad_set_active (ogm->srcpad, FALSE);
       gst_element_remove_pad (GST_ELEMENT (ogm), ogm->srcpad);
       ogm->srcpad = NULL;
     } else {
@@ -639,6 +648,7 @@ gst_ogm_parse_stream_header (GstOgmParse * ogm, const guint8 * data, guint size)
     ogm->srcpad = gst_pad_new_from_template (ogm->srcpadtempl, "src");
     gst_pad_use_fixed_caps (ogm->srcpad);
     gst_pad_set_caps (ogm->srcpad, caps);
+    gst_pad_set_active (ogm->srcpad, TRUE);
     gst_element_add_pad (GST_ELEMENT (ogm), ogm->srcpad);
     GST_INFO_OBJECT (ogm, "Added pad %s:%s with caps %" GST_PTR_FORMAT,
         GST_DEBUG_PAD_NAME (ogm->srcpad), caps);
@@ -655,6 +665,15 @@ gst_ogm_parse_stream_header (GstOgmParse * ogm, const guint8 * data, guint size)
       gst_pad_push_event (ogm->srcpad, event);
     }
     g_list_free (cached_events);
+
+    {
+      GstTagList *tags;
+
+      tags = gst_tag_list_new ();
+      gst_tag_list_add (tags, GST_TAG_MERGE_APPEND, GST_TAG_SUBTITLE_CODEC,
+          "Ogm", NULL);
+      gst_element_found_tags_for_pad (GST_ELEMENT (ogm), ogm->srcpad, tags);
+    }
   }
 
   gst_caps_unref (caps);
@@ -920,6 +939,7 @@ gst_ogm_parse_change_state (GstElement * element, GstStateChange transition)
   switch (transition) {
     case GST_STATE_CHANGE_PAUSED_TO_READY:
       if (ogm->srcpad) {
+        gst_pad_set_active (ogm->srcpad, FALSE);
         gst_element_remove_pad (element, ogm->srcpad);
         ogm->srcpad = NULL;
       }
@@ -939,6 +959,8 @@ gst_ogm_parse_change_state (GstElement * element, GstStateChange transition)
 gboolean
 gst_ogm_parse_plugin_init (GstPlugin * plugin)
 {
+  gst_riff_init ();
+
   GST_DEBUG_CATEGORY_INIT (gst_ogm_parse_debug, "ogmparse", 0, "ogm parser");
 
   return gst_element_register (plugin, "ogmaudioparse", GST_RANK_PRIMARY,
