@@ -9,7 +9,7 @@
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more 
+ * Library General Public License for more
  */
 
 /**
@@ -45,6 +45,10 @@ struct _GstBaseRTPPayloadPrivate
   gboolean seqnum_offset_random;
   gboolean ssrc_random;
   guint16 next_seqnum;
+  gboolean perfect_rtptime;
+
+  gint64 prop_max_ptime;
+  gint64 caps_max_ptime;
 };
 
 /* BaseRTPPayload signals and args */
@@ -54,7 +58,7 @@ enum
   LAST_SIGNAL
 };
 
-/* FIXME 0.11, a better default is the Ethernet MTU of 
+/* FIXME 0.11, a better default is the Ethernet MTU of
  * 1500 - sizeof(headers) as pointed out by marcelm in IRC:
  * So an Ethernet MTU of 1500, minus 60 for the max IP, minus 8 for UDP, gives
  * 1432 bytes or so.  And that should be adjusted downward further for other
@@ -67,6 +71,7 @@ enum
 #define DEFAULT_SEQNUM_OFFSET           -1
 #define DEFAULT_MAX_PTIME               -1
 #define DEFAULT_MIN_PTIME               0
+#define DEFAULT_PERFECT_RTPTIME         TRUE
 
 enum
 {
@@ -79,7 +84,9 @@ enum
   PROP_MAX_PTIME,
   PROP_MIN_PTIME,
   PROP_TIMESTAMP,
-  PROP_SEQNUM
+  PROP_SEQNUM,
+  PROP_PERFECT_RTPTIME,
+  PROP_LAST
 };
 
 static void gst_basertppayload_class_init (GstBaseRTPPayloadClass * klass);
@@ -204,6 +211,21 @@ gst_basertppayload_class_init (GstBaseRTPPayloadClass * klass)
           "The RTP sequence number of the last processed packet",
           0, G_MAXUINT16, 0, G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
 
+  /**
+   * GstBaseRTPAudioPayload:perfect-rtptime:
+   *
+   * Try to use the offset fields to generate perfect RTP timestamps. when this
+   * option is disabled, RTP timestamps are generated from the GStreamer
+   * timestamps, which could result in RTP timestamps that don't increment with
+   * the amount of data in the packet.
+   *
+   * Since: 0.10.25
+   */
+  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_PERFECT_RTPTIME,
+      g_param_spec_boolean ("perfect-rtptime", "Perfect RTP Time",
+          "Generate perfect RTP timestamps when possible",
+          DEFAULT_PERFECT_RTPTIME, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
   gstelement_class->change_state = gst_basertppayload_change_state;
 
   GST_DEBUG_CATEGORY_INIT (basertppayload_debug, "basertppayload", 0,
@@ -241,9 +263,9 @@ gst_basertppayload_init (GstBaseRTPPayload * basertppayload, gpointer g_class)
       gst_basertppayload_chain);
   gst_element_add_pad (GST_ELEMENT (basertppayload), basertppayload->sinkpad);
 
-  basertppayload->seq_rand = g_rand_new ();
-  basertppayload->ssrc_rand = g_rand_new ();
-  basertppayload->ts_rand = g_rand_new ();
+  basertppayload->seq_rand = g_rand_new_with_seed (g_random_int ());
+  basertppayload->ssrc_rand = g_rand_new_with_seed (g_random_int ());
+  basertppayload->ts_rand = g_rand_new_with_seed (g_random_int ());
 
   basertppayload->mtu = DEFAULT_MTU;
   basertppayload->pt = DEFAULT_PT;
@@ -256,11 +278,15 @@ gst_basertppayload_init (GstBaseRTPPayload * basertppayload, gpointer g_class)
 
   basertppayload->max_ptime = DEFAULT_MAX_PTIME;
   basertppayload->min_ptime = DEFAULT_MIN_PTIME;
+  basertppayload->priv->perfect_rtptime = DEFAULT_PERFECT_RTPTIME;
 
   basertppayload->media = NULL;
   basertppayload->encoding_name = NULL;
 
   basertppayload->clock_rate = 0;
+
+  basertppayload->priv->caps_max_ptime = DEFAULT_MAX_PTIME;
+  basertppayload->priv->prop_max_ptime = DEFAULT_MAX_PTIME;
 }
 
 static void
@@ -418,7 +444,7 @@ no_function:
  * @payload: a #GstBaseRTPPayload
  * @media: the media type (typically "audio" or "video")
  * @dynamic: if the payload type is dynamic
- * @encoding_name: the encoding name 
+ * @encoding_name: the encoding name
  * @clock_rate: the clock rate of the media
  *
  * Set the rtp options of the payloader. These options will be set in the caps
@@ -449,6 +475,21 @@ copy_fixed (GQuark field_id, const GValue * value, GstStructure * dest)
   return TRUE;
 }
 
+static void
+update_max_ptime (GstBaseRTPPayload * basertppayload)
+{
+  if (basertppayload->priv->caps_max_ptime != -1 &&
+      basertppayload->priv->prop_max_ptime != -1)
+    basertppayload->max_ptime = MIN (basertppayload->priv->caps_max_ptime,
+        basertppayload->priv->prop_max_ptime);
+  else if (basertppayload->priv->caps_max_ptime != -1)
+    basertppayload->max_ptime = basertppayload->priv->caps_max_ptime;
+  else if (basertppayload->priv->prop_max_ptime != -1)
+    basertppayload->max_ptime = basertppayload->priv->prop_max_ptime;
+  else
+    basertppayload->max_ptime = DEFAULT_MAX_PTIME;
+}
+
 /**
  * gst_basertppayload_set_outcaps:
  * @payload: a #GstBaseRTPPayload
@@ -469,7 +510,7 @@ gst_basertppayload_set_outcaps (GstBaseRTPPayload * payload, gchar * fieldname,
   GstCaps *srccaps, *peercaps;
   gboolean res;
 
-  /* fill in the defaults, there properties cannot be negotiated. */
+  /* fill in the defaults, their properties cannot be negotiated. */
   srccaps = gst_caps_new_simple ("application/x-rtp",
       "media", G_TYPE_STRING, payload->media,
       "clock-rate", G_TYPE_INT, payload->clock_rate,
@@ -488,6 +529,9 @@ gst_basertppayload_set_outcaps (GstBaseRTPPayload * payload, gchar * fieldname,
     GST_DEBUG_OBJECT (payload, "custom added: %" GST_PTR_FORMAT, srccaps);
   }
 
+  payload->priv->caps_max_ptime = DEFAULT_MAX_PTIME;
+  payload->abidata.ABI.ptime = 0;
+
   /* the peer caps can override some of the defaults */
   peercaps = gst_pad_peer_get_caps (payload->srcpad);
   if (peercaps == NULL) {
@@ -504,6 +548,7 @@ gst_basertppayload_set_outcaps (GstBaseRTPPayload * payload, gchar * fieldname,
     GstStructure *s, *d;
     const GValue *value;
     gint pt;
+    guint max_ptime, ptime;
 
     /* peer provides caps we can use to fixate, intersect. This always returns a
      * writable caps. */
@@ -511,12 +556,23 @@ gst_basertppayload_set_outcaps (GstBaseRTPPayload * payload, gchar * fieldname,
     gst_caps_unref (srccaps);
     gst_caps_unref (peercaps);
 
+    if (gst_caps_is_empty (temp)) {
+      gst_caps_unref (temp);
+      return FALSE;
+    }
+
     /* now fixate, start by taking the first caps */
     gst_caps_truncate (temp);
 
     /* get first structure */
     s = gst_caps_get_structure (temp, 0);
 
+    if (gst_structure_get_uint (s, "maxptime", &max_ptime))
+      payload->priv->caps_max_ptime = max_ptime * GST_MSECOND;
+
+    if (gst_structure_get_uint (s, "ptime", &ptime))
+      payload->abidata.ABI.ptime = ptime * GST_MSECOND;
+
     if (gst_structure_get_int (s, "payload", &pt)) {
       /* use peer pt */
       GST_BASE_RTP_PAYLOAD_PT (payload) = pt;
@@ -582,6 +638,8 @@ gst_basertppayload_set_outcaps (GstBaseRTPPayload * payload, gchar * fieldname,
     GST_DEBUG_OBJECT (payload, "with peer caps: %" GST_PTR_FORMAT, srccaps);
   }
 
+  update_max_ptime (payload);
+
   res = gst_pad_set_caps (GST_BASE_RTP_PAYLOAD_SRCPAD (payload), srccaps);
   gst_caps_unref (srccaps);
 
@@ -594,7 +652,7 @@ gst_basertppayload_set_outcaps (GstBaseRTPPayload * payload, gchar * fieldname,
  * @size: the size of the packet
  * @duration: the duration of the packet
  *
- * Check if the packet with @size and @duration would exceed the configure
+ * Check if the packet with @size and @duration would exceed the configured
  * maximum size.
  *
  * Returns: %TRUE if the packet of @size and @duration would exceed the
@@ -613,6 +671,168 @@ gst_basertppayload_is_filled (GstBaseRTPPayload * payload,
   return FALSE;
 }
 
+typedef struct
+{
+  GstBaseRTPPayload *payload;
+  guint32 ssrc;
+  guint16 seqnum;
+  guint8 pt;
+  GstCaps *caps;
+  GstClockTime timestamp;
+  guint64 offset;
+  guint32 rtptime;
+} HeaderData;
+
+static GstBufferListItem
+find_timestamp (GstBuffer ** buffer, guint group, guint idx, HeaderData * data)
+{
+  data->timestamp = GST_BUFFER_TIMESTAMP (*buffer);
+  data->offset = GST_BUFFER_OFFSET (*buffer);
+
+  /* stop when we find a timestamp. We take whatever offset is associated with
+   * the timestamp (if any) to do perfect timestamps when we need to. */
+  if (data->timestamp != -1)
+    return GST_BUFFER_LIST_END;
+  else
+    return GST_BUFFER_LIST_CONTINUE;
+}
+
+static GstBufferListItem
+set_headers (GstBuffer ** buffer, guint group, guint idx, HeaderData * data)
+{
+  gst_rtp_buffer_set_ssrc (*buffer, data->ssrc);
+  gst_rtp_buffer_set_payload_type (*buffer, data->pt);
+  gst_rtp_buffer_set_seq (*buffer, data->seqnum);
+  gst_rtp_buffer_set_timestamp (*buffer, data->rtptime);
+  gst_buffer_set_caps (*buffer, data->caps);
+  /* increment the seqnum for each buffer */
+  data->seqnum++;
+
+  return GST_BUFFER_LIST_SKIP_GROUP;
+}
+
+/* Updates the SSRC, payload type, seqnum and timestamp of the RTP buffer
+ * before the buffer is pushed. */
+static GstFlowReturn
+gst_basertppayload_prepare_push (GstBaseRTPPayload * payload,
+    gpointer obj, gboolean is_list)
+{
+  GstBaseRTPPayloadPrivate *priv;
+  HeaderData data;
+
+  if (payload->clock_rate == 0)
+    goto no_rate;
+
+  priv = payload->priv;
+
+  /* update first, so that the property is set to the last
+   * seqnum pushed */
+  payload->seqnum = priv->next_seqnum;
+
+  /* fill in the fields we want to set on all headers */
+  data.payload = payload;
+  data.seqnum = payload->seqnum;
+  data.ssrc = payload->current_ssrc;
+  data.pt = payload->pt;
+  data.caps = GST_PAD_CAPS (payload->srcpad);
+
+  /* find the first buffer with a timestamp */
+  if (is_list) {
+    data.timestamp = -1;
+    data.offset = GST_BUFFER_OFFSET_NONE;
+    gst_buffer_list_foreach (GST_BUFFER_LIST_CAST (obj),
+        (GstBufferListFunc) find_timestamp, &data);
+  } else {
+    data.timestamp = GST_BUFFER_TIMESTAMP (GST_BUFFER_CAST (obj));
+    data.offset = GST_BUFFER_OFFSET (GST_BUFFER_CAST (obj));
+  }
+
+  /* convert to RTP time */
+  if (priv->perfect_rtptime && data.offset != GST_BUFFER_OFFSET_NONE) {
+    /* if we have an offset, use that for making an RTP timestamp */
+    data.rtptime = payload->ts_base + data.offset;
+    GST_LOG_OBJECT (payload,
+        "Using offset %" G_GUINT64_FORMAT " for RTP timestamp", data.offset);
+  } else if (GST_CLOCK_TIME_IS_VALID (data.timestamp)) {
+    gint64 rtime;
+
+    /* no offset, use the gstreamer timestamp */
+    rtime = gst_segment_to_running_time (&payload->segment, GST_FORMAT_TIME,
+        data.timestamp);
+
+    GST_LOG_OBJECT (payload,
+        "Using running_time %" GST_TIME_FORMAT " for RTP timestamp",
+        GST_TIME_ARGS (rtime));
+
+    rtime = gst_util_uint64_scale_int (rtime, payload->clock_rate, GST_SECOND);
+
+    /* add running_time in clock-rate units to the base timestamp */
+    data.rtptime = payload->ts_base + rtime;
+  } else {
+    GST_LOG_OBJECT (payload,
+        "Using previous RTP timestamp %" G_GUINT32_FORMAT, payload->timestamp);
+    /* no timestamp to convert, take previous timestamp */
+    data.rtptime = payload->timestamp;
+  }
+
+  /* set ssrc, payload type, seq number, caps and rtptime */
+  if (is_list) {
+    gst_buffer_list_foreach (GST_BUFFER_LIST_CAST (obj),
+        (GstBufferListFunc) set_headers, &data);
+  } else {
+    GstBuffer *buf = GST_BUFFER_CAST (obj);
+    set_headers (&buf, 0, 0, &data);
+  }
+
+  priv->next_seqnum = data.seqnum;
+  payload->timestamp = data.rtptime;
+
+  GST_LOG_OBJECT (payload,
+      "Preparing to push packet with size %d, seq=%d, rtptime=%u, timestamp %"
+      GST_TIME_FORMAT, (is_list) ? -1 :
+      GST_BUFFER_SIZE (GST_BUFFER (obj)), payload->seqnum, data.rtptime,
+      GST_TIME_ARGS (data.timestamp));
+
+  return GST_FLOW_OK;
+
+  /* ERRORS */
+no_rate:
+  {
+    GST_ELEMENT_ERROR (payload, STREAM, NOT_IMPLEMENTED, (NULL),
+        ("subclass did not specify clock-rate"));
+    return GST_FLOW_ERROR;
+  }
+}
+
+/**
+ * gst_basertppayload_push_list:
+ * @payload: a #GstBaseRTPPayload
+ * @list: a #GstBufferList
+ *
+ * Push @list to the peer element of the payloader. The SSRC, payload type,
+ * seqnum and timestamp of the RTP buffer will be updated first.
+ *
+ * This function takes ownership of @list.
+ *
+ * Returns: a #GstFlowReturn.
+ *
+ * Since: 0.10.24
+ */
+GstFlowReturn
+gst_basertppayload_push_list (GstBaseRTPPayload * payload, GstBufferList * list)
+{
+  GstFlowReturn res;
+
+  res = gst_basertppayload_prepare_push (payload, list, TRUE);
+
+  if (G_LIKELY (res == GST_FLOW_OK))
+    res = gst_pad_push_list (payload->srcpad, list);
+  else
+    gst_buffer_list_unref (list);
+
+  return res;
+}
+
 /**
  * gst_basertppayload_push:
  * @payload: a #GstBaseRTPPayload
@@ -620,7 +840,7 @@ gst_basertppayload_is_filled (GstBaseRTPPayload * payload,
  *
  * Push @buffer to the peer element of the payloader. The SSRC, payload type,
  * seqnum and timestamp of the RTP buffer will be updated first.
- * 
+ *
  * This function takes ownership of @buffer.
  *
  * Returns: a #GstFlowReturn.
@@ -629,69 +849,15 @@ GstFlowReturn
 gst_basertppayload_push (GstBaseRTPPayload * payload, GstBuffer * buffer)
 {
   GstFlowReturn res;
-  GstClockTime timestamp;
-  guint32 rtptime;
-  GstBaseRTPPayloadPrivate *priv;
 
-  if (payload->clock_rate == 0)
-    goto no_rate;
+  res = gst_basertppayload_prepare_push (payload, buffer, FALSE);
 
-  priv = payload->priv;
-
-  gst_rtp_buffer_set_ssrc (buffer, payload->current_ssrc);
-
-  gst_rtp_buffer_set_payload_type (buffer, payload->pt);
-
-  /* update first, so that the property is set to the last
-   * seqnum pushed */
-  payload->seqnum = priv->next_seqnum;
-  gst_rtp_buffer_set_seq (buffer, payload->seqnum);
-
-  /* can wrap around, which is perfectly fine */
-  priv->next_seqnum++;
-
-  /* add our random offset to the timestamp */
-  rtptime = payload->ts_base;
-
-  timestamp = GST_BUFFER_TIMESTAMP (buffer);
-  if (GST_CLOCK_TIME_IS_VALID (timestamp)) {
-    gint64 rtime;
-
-    rtime = gst_segment_to_running_time (&payload->segment, GST_FORMAT_TIME,
-        timestamp);
-
-    rtime = gst_util_uint64_scale_int (rtime, payload->clock_rate, GST_SECOND);
-
-    /* add running_time in clock-rate units to the base timestamp */
-    rtptime += rtime;
-  } else {
-    /* no timestamp to convert, take previous timestamp */
-    rtptime = payload->timestamp;
-  }
-  gst_rtp_buffer_set_timestamp (buffer, rtptime);
-
-  payload->timestamp = rtptime;
-
-  /* set caps */
-  gst_buffer_set_caps (buffer, GST_PAD_CAPS (payload->srcpad));
-
-  GST_LOG_OBJECT (payload,
-      "Pushing packet size %d, seq=%d, rtptime=%u, timestamp %" GST_TIME_FORMAT,
-      GST_BUFFER_SIZE (buffer), payload->seqnum, rtptime,
-      GST_TIME_ARGS (timestamp));
-
-  res = gst_pad_push (payload->srcpad, buffer);
+  if (G_LIKELY (res == GST_FLOW_OK))
+    res = gst_pad_push (payload->srcpad, buffer);
+  else
+    gst_buffer_unref (buffer);
 
   return res;
-
-  /* ERRORS */
-no_rate:
-  {
-    GST_ELEMENT_ERROR (payload, STREAM, NOT_IMPLEMENTED, (NULL),
-        ("subclass did not specify clock-rate"));
-    gst_buffer_unref (buffer);
-    return GST_FLOW_ERROR;
-  }
 }
 
 static void
@@ -730,11 +896,15 @@ gst_basertppayload_set_property (GObject * object, guint prop_id,
           basertppayload->seqnum_offset, priv->seqnum_offset_random);
       break;
     case PROP_MAX_PTIME:
-      basertppayload->max_ptime = g_value_get_int64 (value);
+      basertppayload->priv->prop_max_ptime = g_value_get_int64 (value);
+      update_max_ptime (basertppayload);
       break;
     case PROP_MIN_PTIME:
       basertppayload->min_ptime = g_value_get_int64 (value);
       break;
+    case PROP_PERFECT_RTPTIME:
+      priv->perfect_rtptime = g_value_get_boolean (value);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -788,6 +958,9 @@ gst_basertppayload_get_property (GObject * object, guint prop_id,
     case PROP_SEQNUM:
       g_value_set_uint (value, basertppayload->seqnum);
       break;
+    case PROP_PERFECT_RTPTIME:
+      g_value_set_boolean (value, priv->perfect_rtptime);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -812,20 +985,19 @@ gst_basertppayload_change_state (GstElement * element,
       gst_segment_init (&basertppayload->segment, GST_FORMAT_UNDEFINED);
 
       if (priv->seqnum_offset_random)
-        basertppayload->seqnum_base =
-            g_rand_int_range (basertppayload->seq_rand, 0, G_MAXUINT16);
+        basertppayload->seqnum_base = g_random_int_range (0, G_MAXUINT16);
       else
         basertppayload->seqnum_base = basertppayload->seqnum_offset;
       priv->next_seqnum = basertppayload->seqnum_base;
       basertppayload->seqnum = basertppayload->seqnum_base;
 
       if (priv->ssrc_random)
-        basertppayload->current_ssrc = g_rand_int (basertppayload->ssrc_rand);
+        basertppayload->current_ssrc = g_random_int ();
       else
         basertppayload->current_ssrc = basertppayload->ssrc;
 
       if (priv->ts_offset_random)
-        basertppayload->ts_base = g_rand_int (basertppayload->ts_rand);
+        basertppayload->ts_base = g_random_int ();
       else
         basertppayload->ts_base = basertppayload->ts_offset;
       basertppayload->timestamp = basertppayload->ts_base;
